;;; "fuelle-ts-entwicklung-2012.scm"
;;; Demonstration der Entwicklung des Programmes zur Tiefensuche am Rucksackproblem
(require (lib "breakpoint.ss"))
; ----- Liste der vorhandenen Stuecke -----
(define stuecke '(30 30 30 20 20 15 5 2 1))

;;; ===== ANFANG HILFSFUNKTIONEN =====
; ----- Praedikate -----
; Fuer das Verstaendnis ist vermutlich einfacher, diese Prädikate auszugliedern:
;;; ----- Hilfsfunktion exakt-voll? -----
; prüft, ob der Container [Rucksack] die Zielgröße erreicht hat.
(define
  (exakt-voll? container zielgroesse)
  (= (apply + container) zielgroesse)
  )

;;; ----- Hilfsfunktion zu-voll? -----
; prüft, ob das Stück noch in den container passt.
(define
  (zu-voll? stueck container zielgroesse)
  (> (+ stueck (apply + container)) zielgroesse))

;;; ===== ENDE HILFSFUNKTIONEN =====

;;; ----- Funktionskopf -----
; 1. Schritt: Bestimmung des Funktionskopfes
(define
  (tiefensuche stuecke container zielgroesse)
  'wert-undefiniert)
; Testaufrufe:
(writeln "Schritt 1: ")
(tiefensuche stuecke () 105)
(tiefensuche stuecke () 100)

;;; ----- exakt gefüllt? -----
; 2. Schritt: ich bin am Ziel, der Container ist exakt gefuellt.
; Hier ist also eine Bedingung zu überprüfen. --> cond
(define
  (tiefensuche stuecke container zielgroesse)
  (cond
    [(exakt-voll? container zielgroesse)       ; Bedingung
     container]                                ; ja-Fall: container zurueckgeben
    [else
     'wert-undefiniert]                        ; nein-Fall
   )
  )

; Testaufrufe:
(writeln "Schritt 2: ")
(tiefensuche '(5 2 1) '(30 30 30 15) 105)
(tiefensuche '(15 5 2 1) '(30 30 30) 100)

;;; ----- am "Ende", aber nicht am Ziel -----
; 3. Schritt: es geht nicht weiter!
; Alle Stuecke wurden versucht, ohne dass sich der Container passend fuellen ließ.
(define
  (tiefensuche stuecke container zielgroesse)
  (cond
    [(exakt-voll? container zielgroesse)       ; Bedingung
     container]                                ; ja-Fall: container zurueckgeben
    [(null? stuecke)                           ; 2.Bedingung
     #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
    [else                                      ; und sonst?
     'wert-undefiniert]                        ; der "sonst"-Fall
    )
  )
; Testaufrufe:
(writeln "Schritt 3: ")
(tiefensuche '(5 2 1) '(30 30 30 15) 105)
(tiefensuche '(5 2 1) '(30 30 30 15) 100)
(tiefensuche '() '(30 30 30 15) 100)

;;; ----- zu voll -----
; 4. Schritt: das aktuelle Stueck passt nicht hinein.
; es wird weitergesucht ohne dieses Stueck.
(define
  (tiefensuche stuecke container zielgroesse)
  (cond
    [(exakt-voll? container zielgroesse)       ; Bedingung
     container]                                ; ja-Fall: container zurueckgeben
    [(null? stuecke)                           ; 2.Bedingung
     #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
    [(zu-voll? (first stuecke) container zielgroesse)  ; prüfen
     #f]                                       ; so geht es nicht. #f kennzeichnet Misserfolg.
    [else                                      ; und sonst?
     'wert-undefiniert]                        ; der "sonst"-Fall
    )
  )

; Testaufrufe:
(writeln "Schritt 4: ")
(tiefensuche stuecke '() 105)
(tiefensuche stuecke '() 100)
(tiefensuche stuecke '() 103)
(tiefensuche stuecke '() 500)


;;; ----- der "sonst"-Fall ??? -----
; 5. Schritt: Es wird weitergesucht mit dem aktuellen Stueck im Container.
(define
  (tiefensuche stuecke container zielgroesse)
  (cond
    [(exakt-voll? container zielgroesse)       ; Bedingung
     container]                                ; ja-Fall: container zurueckgeben
    [(null? stuecke)                           ; 2.Bedingung
     #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
    [(zu-voll? (first stuecke) container zielgroesse)  ; prüfen
     #f]                                       ; so geht es nicht. #f kennzeichnet Misserfolg.
    [else                                      ; und sonst?
     (tiefensuche                              ; weiter suchen:.
      (rest stuecke)                           ; das erste Element wechselt ...
      (cons (first stuecke) container)         ; ... in den Container;
      zielgroesse)]                            ; es bleibt beim selben Zielwert.
    )
  )

; Testaufrufe:
(writeln "Schritt 5: ")
(tiefensuche stuecke '() 105)
(tiefensuche stuecke '() 100)
(tiefensuche stuecke '() 103)
(tiefensuche stuecke '() 500)

;;; ----- das war es noch nicht ! -----
; 6. Schritt: das Stueck einzupacken fuehrte ggf. in eine Sackgasse.
; Wir muessen es auch ohne dies Stueck versuchen.
; Dazu muss der Aufruf mit dem Stueck (das ist der Schritt in die Tiefe)
; wie eine Bedingung behandelt werden. Wie das z.B. geht, s.u.
(define
  (tiefensuche stuecke container zielgroesse)
  (cond
    [(exakt-voll? container zielgroesse)       ; Bedingung
     container]                                ; ja-Fall: container zurueckgeben
    [(null? stuecke)                           ; 2.Bedingung
     #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
    [(zu-voll? (first stuecke) container zielgroesse)  ; prüfen
     #f]                                       ; so geht es nicht. #f kennzeichnet Misserfolg.
     [(tiefensuche                             ; weiter suchen ...
      (rest stuecke)                           ; das erste Element wechselt ...
      (cons (first stuecke) container)         ; ... in den Container;
      zielgroesse)]                            ; es bleibt beim selben Zielwert.
    ;;; Eine der besonderen Moeglichkeiten von cond: Bedingung=Wert
    ;;; Die Bedingung hat einen Wert, der zurückgegeben werden kann.
    ;;; Bei erfolglosem Versuch brauchen wir dann hier den Wert #f,
    ;;; damit zum else gegangen wird.
    [else                                      ; und sonst?
     (tiefensuche                              ; Wir müssen weiter suchen ...
      (rest stuecke)                           ; ... ohne das erste Element ...
      container                                ; ... mit der bisherigen Fuellung
      zielgroesse)]                            ; ... und dem selben Zielwert.
    )
  )

; Testaufrufe:
(writeln "Schritt 6: ")
(tiefensuche stuecke '() 105)
(tiefensuche stuecke '() 100)
(tiefensuche stuecke '() 103)
(tiefensuche stuecke '() 500)


;;; ----- Vervollständigung der Ausgabe -----
; 7. Schritt: kommentierte Ausgabe!
(define
  (tiefensuche stuecke container zielgroesse)
  (cond
    [(exakt-voll? container zielgroesse)       ; 1. Bedingung
     (writeln "Fuellung " zielgroesse " erreicht mit Container " container)
     container]                                ; ja-Fall
    [(null? stuecke)                           ; 2.Bedingung
     #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
    [(zu-voll? (first stuecke) container zielgroesse)  ; prüfen
     #f]                                       ; so geht es nicht. #f kennzeichnet Misserfolg.
     [(tiefensuche                             ; weiter suchen ...
      (rest stuecke)                           ; das erste Element wechselt ...
      (cons (first stuecke) container)         ; ... in den Container;
      zielgroesse)]                            ; es bleibt beim selben Zielwert.
    [else                                      ; und sonst?
     (tiefensuche                              ; Wir müssen weiter suchen ...
      (rest stuecke)                           ; ... ohne das erste Element ...
      container                                ; ... mit der bisherigen Fuellung
      zielgroesse)]                            ; ... und dem selben Zielwert.
    )
  )


; Testaufrufe:
(writeln "Schritt 7: ")
(tiefensuche stuecke '() 105)
(tiefensuche stuecke '() 100)
(tiefensuche stuecke '() 103)
(tiefensuche stuecke '() 500)
